/*
 * This script InsertExplodedBlocks a track with parameters according to the user dialog
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2015-01-18
 */

//http://www.ribbonsoft.com/doc/qcad/3.0/developer/class_r_entity.html

include("../QCTrack.js");
include("../../WidgetFactory.js");

function InsertExplodedBlock(guiAction) {
    QCTrack.call(this, guiAction);
}

InsertExplodedBlock.prototype = new EAction();
InsertExplodedBlock.includeBasePath = includeBasePath;

InsertExplodedBlock.State = {
    SettingDestination : 0,
};

InsertExplodedBlock.prototype.beginEvent = function() {
    EAction.prototype.beginEvent.call(this);
    var document = EAction.getDocument();
    var di = EAction.getDocumentInterface();

    this.blockId = getActiveBlockId();
    if (this.blockId === RObject.INVALID_ID) {
      EAction.handleUserMessage(qsTr("no valid block selected in BlockList"));
      (new QMessageBox(null, qsTr("Insert exploded block"), qsTr("no valid block selected in BlockList, did not insert anything"), QMessageBox.Ok, null, Qt.Dialog)).exec();
      this.terminate();
      return;
    }
    this.setState(InsertExplodedBlock.State.SettingDestination);
};

InsertExplodedBlock.prototype.setState = function(state) {
    EAction.prototype.setState.call(this, state);

    this.getDocumentInterface().setClickMode(RAction.PickCoordinate);
    this.setCrosshairCursor();

    var appWin = RMainWindowQt.getMainWindow();
    var info;
    switch (this.state) {
    case InsertExplodedBlock.State.SettingDestination:
      this.lineDest = undefined;
      info = qsTr("Select destination");
      appWin.setCommandPrompt(info);
      appWin.setLeftMouseTip(info);
      appWin.setRightMouseTip(EAction.trBack);
      break;
    }
};

InsertExplodedBlock.prototype.escapeEvent = function() {
    switch (this.state) {
    case InsertExplodedBlock.State.SettingDestination:
      EAction.prototype.escapeEvent.call(this);
      break;
    }
};

InsertExplodedBlock.prototype.pickCoordinate = function(event, preview) {
    var di = this.getDocumentInterface();
    var document = this.getDocument();
    this.pos = event.getModelPosition();
    var op;

    switch (this.state) {
    case InsertExplodedBlock.State.SettingDestination:
      if (preview) {
        this.updatePreview();
      }
      else {
        op = this.getOperation(false);
        if (op == undefined) {
          break;
        }
        di.applyOperation(op);
        di.setRelativeZero(this.pos);
        this.setState(-1);
        this.terminate();
      }
      break;
    }
};

InsertExplodedBlock.prototype.getOperation = function(preview) {
    var document = this.getDocument();
    var storage = document.getStorage();
    if (isNull(document)) {
      return undefined;
    }
    var currentBlockId = document.getCurrentBlockId();

    var op = new RAddObjectsOperation();
    var subIds = document.queryBlockEntities(this.blockId);
    if (isNull(subIds)) {
      EAction.handleUserMessage("internal error");
      this.setState(-1);
      this.terminate();
      return;
    }
    for (var j = 0; j < subIds.length; j++) {
      var sub = document.queryEntity(subIds[j]);
      var clone = sub.clone();
      storage.setObjectId(clone, RObject.INVALID_ID);
      clone.setBlockId(currentBlockId);
      clone.move(this.pos);
      op.addObject(clone, false);
    }

    return op;
};

//partly taken from Block.js
getActiveBlockId = function() {
    var blockList = EAction.getMainWindow().findChild("BlockList");
    if (isNull(blockList)) {
      return RObject.INVALID_ID;
    }

    var item = blockList.currentItem();
    if (isNull(item)) {
      return RObject.INVALID_ID;
    }

    var name = item.text();
    if (isNull(name) || name.startsWith("*")) {
      return RObject.INVALID_ID;
    }

    var document = EAction.getDocument();
    if (isNull(document)) {
      return RObject.INVALID_ID;
    }


    var blockId = document.getBlockId(name);

    return blockId;
};

InsertExplodedBlock.init = function(basePath) {
    var action = new RGuiAction(QCTrack.MenuItem.InsertExplodedBlock, RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/InsertExplodedBlock.js");
    action.setGroupSortOrder(QCTrack.MenuOrder.GroupSortOrder);
    action.setSortOrder(QCTrack.MenuOrder.InsertExplodedBlock);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
};

