/*
 * This file leads to a menu in QCad and provides common functions.
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2015-01-15
 */

include("../EAction.js");

function QCTrack(guiAction) {
    EAction.call(this, guiAction);
}

QCTrack.prototype = new EAction();
QCTrack.includeBasePath = includeBasePath;

QCTrack.MenuOrder = {
    GroupSortOrder      : 1612340,
    Create              : 1612401,
    InsertExplodedBlock : 1612402,
    EnsureOrder         : 1612403,
    SelectCon           : 1612404,
    Move                : 1612405,
    DupMove             : 1612406,
    DistanceInfo        : 1612407,
    Align               : 1612408,
    FillStraight        : 1612409,
    Variegate           : 1612410,
    TrackWidget         : 1612411,
    TrackWidgetInsert   : 1612412,
    TrackWidgetRotate   : 1612413,
    PartList            : 1612414
};

QCTrack.MenuItem = {
    Create              : qsTr("&Create track"),
    InsertExplodedBlock : qsTr("Insert e&xploded block from BlockListWidget"),
    EnsureOrder         : qsTr("Ensure &order of track layer objects"),
    SelectCon           : qsTr("&Select connected tracks"),
    Move                : qsTr("&Move"),
    DupMove             : qsTr("&Duplicate and move"),
    DistanceInfo        : qsTr("Distance &information"),
    Align               : qsTr("&Align"),
    FillStraight        : qsTr("&Fill with straight tracks"),
    Variegate           : qsTr("&Variegate"),
    TrackWidget         : qsTr("&View track widget"),
    TrackWidgetInsert   : qsTr("Insert &new track from track widget"),
    TrackWidgetRotate   : qsTr("&Rotate track at track widget"),
    PartList            : qsTr("Show &part list")
};

QCTrack.RadiusColors = [
    new QColor(224, 0, 128),
    new QColor(0, 144, 56),
    new QColor(255, 204, 0),
    new QColor(128, 96, 160),
    new QColor(192, 192, 192),
    new QColor(88, 88, 88)
];

QCTrack.getMenu = function() {
    var menu = EAction.getMenu(QCTrack.getTitle(), "QCTrackMenu");
    menu.setProperty("scriptFile", QCTrack.includeBasePath + "/QCTrack.js");
    return menu;
};

QCTrack.getToolBar = function() {
    var tb = EAction.getToolBar(QCTrack.getTitle(), "QCTrackToolBar");
    tb.visible = false;
    return tb;
};

QCTrack.getCadToolBarPanel = function() {
    return EAction.getMainCadToolBarPanel();
};

QCTrack.getTitle = function() {
    return qsTr("&QCTrack");
};

QCTrack.prototype.getTitle = function() {
    return QCTrack.getTitle();
};

// common functions

function checkLayerExistence(document, di, layerName, colorName) {
    if (!document.hasLayer(layerName)) {
      var color = new RColor(colorName);
      var linetypeContinuousId = document.getLinetypeId("continuous");
      var myLayer = new RLayer(document, layerName, false, false, color, linetypeContinuousId, RLineweight.Weight025);
      var opAddLayer = new RAddObjectOperation(myLayer);
      di.applyOperation(opAddLayer);
    }
    else {
      var op = new RModifyObjectsOperation();
      var layer = document.queryLayer(layerName);
      layer.setFrozen(false);
      layer.setLocked(false);
      op.addObject(layer);
      di.applyOperation(op);
    }
}

function addRectangle(document, operation, p1, p2, p3, p4) {
    var line;
    line = new RLineEntity(document, new RLineData(p1, p2));
    operation.addObject(line);
    line = new RLineEntity(document, new RLineData(p2, p3));
    operation.addObject(line);
    line = new RLineEntity(document, new RLineData(p3, p4));
    operation.addObject(line);
    line = new RLineEntity(document, new RLineData(p4, p1));
    operation.addObject(line);
}

function addHatch(document, operation, p1, p2, p3, p4) {
    var data = new RHatchData(true, 0, 0, "");
    data.newLoop();
    data.addBoundary(new RLine(p1, p2));
    data.addBoundary(new RLine(p2, p3));
    data.addBoundary(new RLine(p3, p4));
    data.addBoundary(new RLine(p4, p1));
    var hatch = new RHatchEntity(document, data);
    operation.addObject(hatch);
}

function getClosestTrackEnd(document, block, pos, onlySelected) {
    if (!isBlockReferenceEntity(block)) {
      return undefined;
    }
    var data = block.getData();
    var subIds = document.queryBlockEntities(data.getReferencedBlockId());
    var k, bestDist = -1, bestEnd = undefined;
    for (k = 0; k < subIds.length; k++) {
      var line = data.queryEntity(subIds[k]);
      if (!isLineEntity(line) || line.getLayerName() != "TR_ENDS" || (onlySelected && !line.isSelected())) {
        continue;
      }
      var curDist = pos.getDistanceTo(line.getClosestPointOnEntity(pos));
      if (bestDist == -1 || curDist < bestDist) {
        bestDist = curDist;
        bestEnd = line;
      }
    }
    return bestEnd;
}

function addAuxPreviewObjects(lineEntity, ret) {
    var mid = getPointOnLine(lineEntity, 0.5);
    ret.push(new RCircle(mid, lineEntity.getLength() * 1.5));
    var top = getPointOnLine(lineEntity, 2);
    var bottom = getPointOnLine(lineEntity, -1);
    var out = getPointOnLine(lineEntity, 2).rotate(Math.PI / 2, mid);
    ret.push(new RLine(top, bottom));
    ret.push(new RLine(bottom, out));
    ret.push(new RLine(out, top));
}

function getPointOnLine(lineEntity, alpha) {
    var start = lineEntity.getStartPoint();
    var end = lineEntity.getEndPoint();
    return new RVector(start.getX() * alpha + end.getX() * (1 - alpha), start.getY() * alpha + end.getY() * (1 - alpha));
}

function getNormalizedAngle(a, limit, epsilon) {
    a = (Math.round(a / epsilon) % Math.round(limit / epsilon)) * epsilon;
    if (a < 0) {
      a += limit;
    }
    return a;
}

function generateTrackPreview(blockDoc, blockId, curveColors, iconSize) { //curveColors has to be an Array if radius dependent colors should be generated
    var ms = new RMemoryStorage();
    var si = new RSpatialIndexNavel();
    var doc = new RDocument(ms, si);
    var di = new RDocumentInterface(doc);
    di.setNotifyListeners(false);
    var scene = new RGraphicsSceneQt(di);
    var view = new RGraphicsViewImage();
    view.setAntialiasing(true);
    view.setPaintOrigin(false);
    view.setBackgroundColor(new QColor(Qt.white));
    view.setColorCorrectionOverride(true);
    view.setMargin(2);
    view.setScene(scene);
    view.resizeImage(iconSize, iconSize);

    var operation = undefined;
    var op = new RAddObjectsOperation();
    var subIds;
    subIds = blockDoc.queryBlockEntities(blockId);
    var radius = 0;
    for (var j = 0; j < subIds.length; j++) {
      var sub = blockDoc.queryEntity(subIds[j]);
      if (sub.getLayerName() == "TR_MIDS") {
        if (isArcEntity(sub)) {
          if (radius == 0) {
            radius = sub.getRadius();
          }
          else if (radius != sub.getRadius()) {
            radius = -1;
          }
        }
      }
      else if (sub.getLayerName() == "TR_PLAN") {
        if (isLineEntity(sub)) {
          op.addObject(new RLineEntity(doc, new RLineData(sub.getStartPoint(), sub.getEndPoint())));
        }
        else if (isArcEntity(sub)) {
          op.addObject(new RArcEntity(doc, new RArcData(sub.getCenter(), sub.getRadius(), sub.getStartAngle(), sub.getEndAngle())));
        }
      }
    }
    if (radius == 0) {
      view.setBackgroundColor(new QColor(0, 0, 127));
    }
    else if (radius > 0) {
      var color = undefined;
      if (curveColors != null) {
        for (var i = 0; i < curveColors.length; i++) {
          if (curveColors[i][0] == radius) { //first element: radius
            color = curveColors[i][1];       //second element: QColor
            break;
          }
        }
        if (color == undefined) { //radius not yet existing
          color = QCTrack.RadiusColors[curveColors.length];
          if (color == undefined) {
            color = new QColor(Math.floor(Math.random()*255),Math.floor(Math.random()*255), Math.floor(Math.random()*255)); 
          }
          curveColors.push([radius, color]);
        }
      }
      else {
        color = new QColor(0, 127, 0);
      }
      view.setBackgroundColor(color);
    }
    di.applyOperation(op);
    di.regenerateScenes();
    di.autoZoom();

    return QPixmap.fromImage(view.getBuffer());
}

//get a filename to save something, filedesc "Export", filedescshort "CSV", fileext ".csv"
function getFileNameToSaveFromUser(filedesc, filedescshort, fileext) { //fileext must start with ".", e.g. ".csv"
    var fi;
    var fileName;
    var initialPath = "";
    var drawingFileName = EAction.getDocument().getFileName();
    if (drawingFileName.length !== 0) {
      fi = new QFileInfo(QDir.homePath());
      initialPath = fi.absoluteFilePath() + QDir.separator
              + stripDirtyFlag(EAction.getMdiChild().windowTitle) + fileext;
    } else {
      fi = new QFileInfo(drawingFileName);
      initialPath = fi.path() + QDir.separator + fi.completeBaseName() + fileext;
    }
    var appWin = EAction.getMainWindow();
    var ret = QCTrack.getSaveFileName(appWin, filedesc + " (" + filedescshort + ")",
                                   initialPath, [ filedescshort + " (*" + fileext + ")" ]);
    if (isNull(ret)) {
        return undefined;
    }
    fileName = ret[0];

    if (fileName === "") {
        return undefined;
    }

    return fileName;
}

/**
 * Advanced file save as dialog with extension completion.
 * This is what QFileDialog.getSaveFileName should be.
 * Taken from QCad scripts/File/File.js to avoid import dependencies.
 *
 * \param parentWidget Parent widget or null
 * \param caption Dialog caption
 * \param dir Initial directory of the dialog
 * \param fileName Initial file name to suggest to user
 * \param filterStrings Array of filter strings in the format 'My File Type (*.mft *.mftype)'
 *
 * \return Array with complete file path and selected name filter or undefined.
 */
QCTrack.getSaveFileName = function(parentWidget, caption, dir, filterStrings) {
    var fileDialog = new QFileDialog(parentWidget);
    
    // use native dialog:
    fileDialog.setOption(QFileDialog.DontUseNativeDialog, false);

    // overwrite is handled after extension has been added:
    fileDialog.setOption(QFileDialog.DontConfirmOverwrite, true);

    var fiDir = new QFileInfo(dir);

    fileDialog.setDirectory(fiDir.absolutePath());
    fileDialog.fileMode = QFileDialog.AnyFile;
    fileDialog.acceptMode = QFileDialog.AcceptSave;

    fileDialog.setNameFilters(filterStrings);
    
    //if (!isNull(fileName)) {
        //var fileInfo = new QFileInfo(fileName);
        fileDialog.selectFile(fiDir.completeBaseName());

        if (fiDir.suffix().length!==0) {
            // preselect first name filter that matches current extension:
            for (var i=0; i<filterStrings.length; ++i) {
                if (filterStrings[i].contains("*." + fiDir.suffix().toLowerCase())) {
                    fileDialog.selectNameFilter(filterStrings[i]);
                    break;
                }
            }
        }
    //}

    fileDialog.setLabelText(QFileDialog.FileType, qsTr("Format:"));

    var done = false;

    while (!done) {
        done = true;

        if (!fileDialog.exec()) {
            fileDialog.destroy();
            return undefined;
        }

        var fileToSave = fileDialog.selectedFiles()[0];
        var selectedNameFilter = fileDialog.selectedNameFilter();

        // file, possibly without suffix:
        var fi = new QFileInfo(fileToSave);
        if (fi.suffix().length===0) {
            var suffix = "";
            var regex = /\(\*(\.[^ )]*)/i;
            if (regex.test(selectedNameFilter)) {
                var matches = selectedNameFilter.match(regex);
                if (matches.length==2) {
                    suffix = selectedNameFilter.match(regex)[1];
                }
            }
            if (suffix.length!==0) {
                fileToSave += suffix;
            }
        }

        // file with suffix:
        fi = new QFileInfo(fileToSave);
        if (fi.exists()) {
            var buttons = new QMessageBox.StandardButtons(QMessageBox.Yes, QMessageBox.No);
            var ret = QMessageBox.warning(parentWidget, 
                qsTr("Overwrite File?"), 
                qsTr("The file '%1' already exists. Do you wish to overwrite it?").arg(fileToSave),
                buttons);
            if (ret!=QMessageBox.Yes) {
                done = false;
            }
        }
    }
    
    return [ fileToSave, selectedNameFilter ];
};

