/*
 * This script inserts the currently selected track
 * from the track widget.
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2015-01-08
 */

include("../QCTrack.js");

function TrackWidgetInsert(guiAction) {
    EAction.call(this, guiAction);
}

TrackWidgetInsert.prototype = new EAction();

TrackWidgetInsert.includeBasePath = includeBasePath;

TrackWidgetInsert.State = {
    SettingLineDest : 0
};

TrackWidgetInsert.prototype.beginEvent = function() {
    this.tracklist = EAction.getMainWindow().findChild("TrackList");
    this.connEnd = EAction.getMainWindow().findChild("connEnd");
    if (isNull(this.tracklist) || isNull(this.connEnd)) {
      this.setState(-1);
      this.terminate();
      return;
    }
    this.item = this.tracklist.currentItem();
    if (isNull(this.item)) {
      this.setState(-1);
      this.terminate();
      return;
    }

    var document = EAction.getDocument();
    if (isNull(document)) {
      this.setState(-1);
      this.terminate();
      return;
    }

    this.refreshBlockInfo(document, true);
    this.setState(TrackWidgetInsert.State.SettingLineDest);
};

TrackWidgetInsert.prototype.refreshBlockInfo = function(document, always) {
    if (!always && (this.item == this.tracklist.currentItem() || isNull(this.tracklist.currentItem()))) {
      return;
    }
    this.item = this.tracklist.currentItem();
    var blockName = this.item.toolTip();
    var block = document.queryBlock(blockName);
    if (isNull(block) || blockName[0] == '*') {
      this.setState(-1);
      this.terminate();
      return;
    }
    this.blockId = block.getId();
    var subIds;
    subIds = document.queryBlockEntities(this.blockId);
    this.blockEnds = new Array();
    for (var j = 0; j < subIds.length; j++) {
      var sub = document.queryEntity(subIds[j]);
      if (sub.getLayerName() == "TR_ENDS") {
        this.blockEnds.push(sub);
      }
    }
    if (this.blockEnds.length == 0) {
      this.setState(-1);
      this.terminate();
      return;
    }
};

TrackWidgetInsert.prototype.setState = function(state) {
    EAction.prototype.setState.call(this, state);

    this.getDocumentInterface().setClickMode(RAction.PickEntity);
    this.setCrosshairCursor();

    var appWin = RMainWindowQt.getMainWindow();
    var info;
    switch (this.state) {
    case TrackWidgetInsert.State.SettingLineDest:
      this.lineDest = undefined;
      info = qsTr("Select destination track end");
      appWin.setCommandPrompt(info);
      appWin.setLeftMouseTip(info);
      appWin.setRightMouseTip(EAction.trCancel);
      break;
    }
};

TrackWidgetInsert.prototype.escapeEvent = function() {
    switch (this.state) {
    case TrackWidgetInsert.State.SettingLineDest:
      EAction.prototype.escapeEvent.call(this);
      break;
    }
};

TrackWidgetInsert.prototype.pickEntity = function(event, preview) {
    var di = this.getDocumentInterface();
    var document = this.getDocument();
    var entityId = event.getEntityId();
    var entity = document.queryEntity(entityId);
    var pos = event.getModelPosition();
    var op;

    switch (this.state) {
    case TrackWidgetInsert.State.SettingLineDest:
      if (isNull(entity)) {
        this.lineDest = undefined;
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      this.lineDest = getClosestTrackEnd(document, entity, pos, false);
      if (this.lineDest == undefined) {
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      if (preview) {
        this.updatePreview();
      }
      else {
        op = this.getOperation(false);
        if (op == undefined) {
          break;
        }
        di.applyOperation(op);
        if (this.blockEnds.length < 2) {
          this.setState(-1);
          this.terminate();
        }
      }
      break;
    }
};

TrackWidgetInsert.prototype.getOperation = function(preview) {
    var document = this.getDocument();
    if (isNull(document) || !isLineEntity(this.lineDest)) {
      return undefined;
    }
    this.refreshBlockInfo(document, false);
    var op = new RAddObjectsOperation();
    var blockReferenceData = new RBlockReferenceData();
    blockReferenceData.setReferencedBlockId(this.blockId);
    var lineSource = this.blockEnds[this.connEnd.currentIndex];
    var dest = this.lineDest.getMiddlePoints()[0], source = lineSource.getMiddlePoints()[0];
    var offset = new RVector(-source.getX(), -source.getY());
    blockReferenceData.setPosition(offset);
    blockReferenceData.setScaleFactors(new RVector(1.0, 1.0));
    blockReferenceData.setRotation(0);
    var rotation = RMath.getNormalizedAngle(this.lineDest.getAngle() - lineSource.getAngle() + Math.PI);
    blockReferenceData.rotate(rotation, new RVector(0, 0));
    blockReferenceData.move(dest);
    var blockRef = new RBlockReferenceEntity(document, blockReferenceData);
    op.addObject(blockRef);
    return op;
};

TrackWidgetInsert.prototype.getAuxPreview = function() {
    var document = this.getDocument();
    this.refreshBlockInfo(document, false);
    var ret = undefined;
    if (isLineEntity(this.lineDest)) {
      //track end connector
      ret = new Array();
      addAuxPreviewObjects(this.lineDest, ret);
      //mid lines
      var lineSource = this.blockEnds[this.connEnd.currentIndex];
      var dest = this.lineDest.getMiddlePoints()[0], source = lineSource.getMiddlePoints()[0];
      var offset = new RVector(-source.getX(), -source.getY());
      var rotation = RMath.getNormalizedAngle(this.lineDest.getAngle() - lineSource.getAngle() + Math.PI);
      var zero = new RVector(0, 0);
      var subIds;
      subIds = document.queryBlockEntities(this.blockId);
      for (var j = 0; j < subIds.length; j++) {
        var sub = document.queryEntity(subIds[j]);
        if (sub.getLayerName() == "TR_MIDS" || sub.getLayerName() == "TR_PLAN") {
          var shape = undefined;
          if (isLineEntity(sub)) {
            shape = new RLine(sub.getStartPoint(), sub.getEndPoint());
          }
          else if (isArcEntity(sub)) {
            shape = new RArc(sub.getCenter(), sub.getRadius(), sub.getStartAngle(), sub.getEndAngle());
          }
          if (shape != undefined) {
            shape.move(offset);
            shape.rotate(rotation, zero);
            shape.move(dest);
            ret.push(shape);
          }
        }
      }
    }
    return ret;
};

TrackWidgetInsert.init = function(basePath) {
    var action = new RGuiAction(QCTrack.MenuItem.TrackWidgetInsert, RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/TrackWidgetInsert.js");
    action.setIcon(basePath + "/TrackWidgetInsert.svg");
    action.setDefaultShortcut(new QKeySequence("q,n"));
    action.setGroupSortOrder(QCTrack.MenuOrder.GroupSortOrder);
    action.setSortOrder(QCTrack.MenuOrder.TrackWidgetInsert);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
    var appWin = EAction.getMainWindow();
    appWin.addSelectionListener(action);
};

