/*
 * This script variegates track combinations with parameters
 * according to the user dialog.
 * Attention! This script is highly re-used by FillStraight, don't
 * change it without checking FillStraight!
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2015-01-15
 */

//http://www.ribbonsoft.com/doc/qcad/3.0/developer/class_r_entity.html

include("../QCTrack.js");
include("../../WidgetFactory.js");

function Variegate(guiAction) {
    QCTrack.call(this, guiAction);
}

Variegate.prototype = new EAction();
Variegate.includeBasePath = includeBasePath;

Variegate.State = {
    SettingPosition : 0
};

Variegate.prototype.beginEvent = function() {
    EAction.prototype.beginEvent.call(this);

    if (!this.doVariegation(undefined)) {
      this.terminate();
      return;
    }

    //give control back to the user
    EAction.handleUserMessage("select position where to place " + this.created + " combinations");
    this.setState(Variegate.State.SettingPosition);
    //do not terminate, user has to set the target point
};

Variegate.prototype.doVariegation = function(destinationLine) {
    var document = EAction.getDocument();

    //get a list of all (straight) tracks
    this.list = new Array(); //each element contains a five-field array containing 0: the block name, 1: the block id, 2: max. usage, 3: its length, and 4: the amount
    var allIds = document.queryAllBlocks();
    var canGenerate = true;
    this.objDistBase = 0;
    for (var i = 0; i < allIds.length; i++) { //first run: collect blocks with exactly one straight line in TR_MIDS
      var id = allIds[i];
      var block = document.queryBlockDirect(id);
      var blockName = block.getName();
      if (blockName[0] == '*') { //reserved blocks like paper space
        continue;
      }
      var k = 0;
      while (k < this.list.length) {
        if (this.list[k][0] == blockName) { //block already existing
          break;
        }
        k++;
      }
      if (k == this.list.length) {
        var subIds;
        subIds = document.queryBlockEntities(id);
        var blockLen = 0;
        var blockHeight = 0;
        var blockEnds = 0;
        var thisCanNotGenerate = false;
        for (var j = 0; j < subIds.length; j++) {
          var sub = document.queryEntity(subIds[j]);
          if (sub.getLayerName() == "TR_MIDS") {
            if (blockLen > 0) { //block has another entity on TR_MIDS layer
              blockLen = 0;
              //EAction.handleUserMessage("block " + blockName + " contains more than one mid element, ignoring it");
              break;
            }
            if (isLineEntity(sub)) {
              blockLen = sub.getStartPoint().getDistanceTo(sub.getEndPoint());
              if (sub.getAngle() != 0) { //can only generate horizontal tracks
                //EAction.handleUserMessage("line in block " + blockName + " has angle != 0, can only print (will not generate tracks)");
                thisCanNotGenerate = true;
              }
            }
            else { //not (just) a straight line
              break;
            }
          }
          else {
            var points = sub.getEndPoints();
            for (var k = 0; k < points.length; k++) {
              var ay = Math.abs(points[k].getY());
              if (ay > blockHeight) {
                blockHeight = ay;
              }
            }
            if (sub.getLayerName() == "TR_ENDS") {
              blockEnds++;
            }
          }
        }
        if (blockLen > 0 && blockEnds == 2) {
          if (thisCanNotGenerate) {
            canGenerate = false;
          }
          this.list.push([blockName, id, 0, blockLen, 0]);
          if (blockHeight > this.objDistBase) {
            this.objDistBase = blockHeight;
          }
        }
      }
    }
    if (this.list.length == 0) {
      EAction.handleUserMessage("no straight line for variegation available");
      return false;
    }
    this.list.sort(function(a, b) {
      return a[0].localeCompare(b[0]);
    });

    //let the user enter the desired parameters
    var dialog = WidgetFactory.createWidget(
                   "scripts/QCTrack/Variegate",
                   "VariegateDialog.ui");
    var widgets = getWidgets(dialog);
    var dialist = widgets["list"];
    var maxSpins = new Array();
    for (var i = 0; i < this.list.length; i++) {
      var w = new QWidget();
      var vbl = new QHBoxLayout(w);
      var item;
      vbl.addWidget(new QLabel(this.list[i][0]), 0, 0);
      vbl.addWidget(item = new QLabel(this.list[i][3].toFixed(1)), 0, 0); 
      item.alignment = Qt.AlignVCenter | Qt.AlignRight;
      vbl.addWidget(item = new QSpinBox(), 0, 0);
      item.minimum = -1;
      item.maximum = 999;
      item.value = 0;
      item.alignment = Qt.AlignVCenter | Qt.AlignRight;
      maxSpins.push(item);
      w.setLayout(vbl);
      var lwi = new QListWidgetItem(dialist);
      lwi.setSizeHint(w.sizeHint);
      dialist.addItem(lwi);
      dialist.setItemWidget(lwi, w);
    }
    var selected = document.querySelectedEntities();
    if (selected.length > 0) {
      for (var i = 0; i < selected.length; i++) {
        var blockName = undefined;
        var id = selected[i];
        var block = document.queryEntity(id);
        if (!isBlockReferenceEntity(block)) {
          continue;
        }
        blockName = block.getReferencedBlockName();
        if (blockName == undefined) {
          continue;
        }
        for (var j = 0; j < this.list.length; j++) {
          if (this.list[j][0] == blockName) {
            maxSpins[j].value = -1;
          }
        }
      }
    }
    else {
      for (var i = 0; i < this.list.length; i++) {
        maxSpins[i].value = -1;
      }
    }
    widgets["useNone"]["clicked()"].connect(this, function() {
      for (var i = 0; i < this.list.length; i++) {
        maxSpins[i].value = 0;
      }
    });
    widgets["useAll"]["clicked()"].connect(this, function() {
      for (var i = 0; i < this.list.length; i++) {
        maxSpins[i].value = -1;
      }
    });
    if (destinationLine != undefined) {
      var destLen = destinationLine.getLength();
      widgets["minLen"].setText((destLen > 5 ? destLen - 5 : 0).toFixed(2));
      widgets["desiredLen"].setText(destLen.toFixed(2));
      widgets["maxLen"].setText((destLen + 5).toFixed(2));
      widgets["optimizeDiffLen"].setText(0.05);
    }
    if (!dialog.exec()) { //the user aborted
      dialog.destroy();
      return false;
    }
    for (var i = this.list.length - 1; i >= 0; i--) {
      if (maxSpins[i].value == 0) {
        this.list.splice(i, 1);
        maxSpins.splice(i, 1);
      }
      else {
        this.list[i][2] = maxSpins[i].value;
      }
    }
    if (this.list.length == 0) { //the user deselected all tracks
      EAction.handleUserMessage("all tracks were disabled by user, can not variegate");
      dialog.destroy();
      return false;
    }
    this.minLen = widgets["minLen"].getValue();
    this.desiredLen = widgets["desiredLen"].getValue();
    this.maxLen = widgets["maxLen"].getValue();
    this.optimizeDiffLen = widgets["optimizeDiffLen"].getValue();
    this.optimizeDiffPerElem = widgets["optimizeDiffPerElem"].getValue();
    this.optimizeDiffRel = widgets["optimizeDiffRel"].getValue() / 100;
    dialog.destroy();

    //produce combinations with backtracking
    EAction.handleUserMessage("variegating with " + this.list.length + " tracks...");
    this.solutions = new Array; //will contain all arrays with [0]: array containing length and element track and [1]: solution arrays particular counts
    this.extendCurrentCombination(0); //do all the work
    if (this.optimizeDiffLen > 0) {
      this.optimize(function(a, b) {
        return compareLength(Math.abs(a[0] - b[0]), this.optimizeDiffLen) < 0;
      });
    }
    if (this.optimizePerElem > 0) {
      this.optimize(function(a, b) {
        return compareLength(Math.abs(a[0] - b[0]), this.optimizeDiffRel * elems) < 0;
      });
    }
    if (this.optimizeDiffRel > 0) {
      this.optimize(function(a, b) {
        return compareLength(Math.abs(a[0] - b[0]), this.optimizeDiffRel * Math.max(a[1], b[1])) < 0;
      });
    }
    this.solutions.sort(function(a, b) {
      var lenDiff = compareLength(a[0][0], b[0][0]); 
      if (lenDiff != 0) {
        return lenDiff;
      }
      return a[0][1] - b[0][1]; //if length is equal, compare element count
    });

    if (this.solutions.length == 0) {
      EAction.handleUserMessage("sorry, did not found a solution");
      this.terminate();
    }
    var bestInd = 0;
    for (var i = 1; i < this.solutions.length; i++) {
      var lenDiff = compareLength(Math.abs(this.desiredLen - this.solutions[i][0][0]), Math.abs(this.desiredLen - this.solutions[bestInd][0][0])); //compare length
      if (lenDiff < 0 || (lenDiff == 0 && this.solutions[i][0][1] < this.solutions[bestInd][0][1])) { //if length is equal, compare element count
        bestInd = i;
      }
    }

    //show results and get further action
    var dialog = WidgetFactory.createWidget(
                   "scripts/QCTrack/Variegate",
                   "VariegateResultDialog.ui");
    var widgets = getWidgets(dialog);
    widgets["resultLabel"].text = "Select solutions (searched: " + this.desiredLen.toFixed(1) + ")";
    this.restable = widgets["resultTable"];
    this.restable.columnCount = this.list.length;
    this.restable.rowCount = this.solutions.length;
    for (var i = 0; i < this.list.length; i++) {
      this.restable.setHorizontalHeaderItem(i, new QTableWidgetItem(this.list[i][0]));
    }
    for (var i = 0; i < this.solutions.length; i++) {
      var item;
      this.restable.setVerticalHeaderItem(i, item = new QTableWidgetItem(this.solutions[i][0][0].toFixed(1) + " / " + this.solutions[i][0][1]));
      item.setTextAlignment(Qt.AlignVCenter | Qt.AlignRight);
      for (var ti = 0; ti < this.list.length; ti++) {
        this.restable.setItem(i, ti, item = new QTableWidgetItem(this.solutions[i][1][ti].toFixed(0)));
        item.setTextAlignment(Qt.AlignVCenter | Qt.AlignRight);
      }
    }
    this.restable.selectRow(bestInd);
    widgets["exportSelected"]["clicked()"].connect(this, function() {
      this.solutionToText(false, false);
    });
    widgets["exportAll"]["clicked()"].connect(this, function() {
      this.solutionToText(false, true);
    });
    widgets["printSelected"]["clicked()"].connect(this, function() {
      this.solutionToText(true, false);
    });
    widgets["printAll"]["clicked()"].connect(this, function() {
      this.solutionToText(true, true);
    });
    widgets["orthoDist"].setText((this.objDistBase * 4).toFixed(2));
    if (destinationLine != undefined) {
      widgets["resultTable"].selectionMode = QAbstractItemView.SingleSelection;
      widgets["orthoDistLabel"].enabled = false;
      widgets["orthoDist"].enabled = false;
      widgets["labelTracks"].enabled = false;
    }
    if (!dialog.exec()) { //the user aborted
      dialog.destroy();
      return false;
    }

    //ok, the user wants us to build tracks in QCad, create the object arrays
    this.createLabels = new Array(); //will contain all label-information with [0]: x, [1]: y, [2]: label text
    this.createBlocks = new Array(); //will contain all block-information with [0]; x, [1]; y, [2]: block id
    this.created = 0;
    var rows = this.restable.selectionModel().selectedRows();
    var distance = widgets["orthoDist"].getValue();
    var doLabel = widgets["labelTracks"].checked;
    for (var i = 0; i < this.solutions.length; i++) {
      if (rows == undefined || isInRowSelection(rows, i)) {
        var y = -distance * this.created;
        var solPck = this.solutions[i];
        if (doLabel) {
          this.createLabels.push([-distance / 4, y, solPck[0][0].toFixed(1)]);
        }
        this.pushBlocksOfSolution(0, y, solPck[1]);
        this.created++;
      }
    }
    return true;
};

Variegate.prototype.solutionToText = function(onlyPrint, all) {
    var rows = undefined;
    if (!all) {
      rows = this.restable.selectionModel().selectedRows();
    }
    if (onlyPrint) { //directly print to QCad console
      for (var i = 0; i < this.solutions.length; i++) {
        if (rows == undefined || isInRowSelection(rows, i)) {
          var solPck = this.solutions[i];
          EAction.handleUserMessage("one solution has " + solPck[0][1]
            + " elements coming to length " + solPck[0][0].toFixed(1) + ":");
          this.printSolution(solPck[1]);
        }
      }
    }
    else { //export to clipboard
      var result = "len,elem";
      for (var i = 0; i < this.list.length; i++) {
        result += "," + this.list[i][0];
      }
      result += "\n";
      for (var i = 0; i < this.solutions.length; i++) {
        if (rows == undefined || isInRowSelection(rows, i)) {
          var solPck = this.solutions[i];
          result += solPck[0][0].toFixed(1) + "," + solPck[0][1];
          for (var j = 0; j < this.list.length; j++) {
            result += "," + solPck[1][j];
          }
          result += "\n";
        }
      }
      getClipboard().setText(result);
    }
};

function isInRowSelection(rows, index) {
    for (var i = 0; i < rows.length; i++) {
      if (rows[i].row() == index) {
        return true;
      }
    }
    return false;
}

Variegate.prototype.optimize = function(testReplacement) {
    var i = 0;
    while (i < this.solutions.length - 1) {
      var j = i + 1;
      var stayHere = false;
      while (j < this.solutions.length) {
        if (testReplacement(this.solutions[i][0], this.solutions[j][0])) {
          if (this.solutions[i][0][1] < this.solutions[j][0][1]) { //i has less elemets
            this.solutions.splice(j, 1);
          }
          else if (this.solutions[i][0][1] > this.solutions[j][0][1]) { //j has less elements
            this.solutions.splice(i, 1);
            stayHere = true;
            break;
          }
          else j++; //same length and same element count, ignore
        }
        else { //different length
          j++;
        }
      }
      if (!stayHere) {
         i++;
      }
    }
}

function compareLength(a, b) { //compare length reducing floating point problems
    return Math.round(a * 2000) - Math.round(b * 2000);
}

Variegate.prototype.extendCurrentCombination = function(startAt) {
    var curLen = this.getCurrentLength();
    var curElems = this.getCurrentElementCount();
    if (curLen > this.maxLen) { //exceeded maximum length or maximum element count
      return;
    }
    if (startAt < this.list.length) {
      while ((this.list[startAt][2] == -1 || this.list[startAt][4] <= this.list[startAt][2]) && curLen <= this.maxLen) { //maximum element count and maximum length must not be exceeded
        this.extendCurrentCombination(startAt + 1);
        this.list[startAt][4]++;
        curLen = this.getCurrentLength();
        curElems = this.getCurrentElementCount();
      }
      this.list[startAt][4]=0; //reset for further back-tracking
    }
    else { //all elements done (maybe some are ==0), check if solution found
      if (curLen >= this.minLen) { //found a solution
        var sol = this.getSolutionArray();
        var len = this.getCombinationLength(sol);
        var elem = this.getCombinationElementCount(sol);
        this.solutions.push([[len, elem], sol]);
      }
    }
};

Variegate.prototype.getCurrentLength = function() {
    var len = 0;
    for (var i = 0; i < this.list.length; i++) {
      len += this.list[i][3] * this.list[i][4];
    }
    return len;
};

Variegate.prototype.getCurrentElementCount = function() {
    var elems = 0;
    for (var i = 0; i < this.list.length; i++) {
      elems += this.list[i][4];
    }
    return elems;
};

Variegate.prototype.getCombinationLength = function(sol) {
    var len = 0;
    for (var i = 0; i < this.list.length; i++) {
      len += this.list[i][3] * sol[i];
    }
    return len;
};

Variegate.prototype.getCombinationElementCount = function(sol) {
    var elems = 0;
    for (var i = 0; i < this.list.length; i++) {
      elems += sol[i];
    }
    return elems;
};

Variegate.prototype.printSolution = function(sol) {
    for (var i = 0; i < this.list.length; i++) {
      EAction.handleUserMessage("-- " + this.list[i][0] + ": " + this.list[i][3] + " * " + sol[i] + " => " + (this.list[i][3] * sol[i]).toFixed(1));
    }
};

Variegate.prototype.pushBlocksOfSolution = function(x, y, sol) {
    for (var i = 0; i < this.list.length; i++) {
      var cnt = sol[i];
      var blockId = this.list[i][1];
      var len = this.list[i][3];
      while (cnt > 0) {
        this.createBlocks.push([x, y, blockId]);
        x += len;
        cnt--;
      }
    }
};

Variegate.prototype.getSolutionArray = function() {
    var sol = new Array(this.list.length);
    for (var i = 0; i < this.list.length; i++) {
      sol[i] = this.list[i][4];
    }
    return sol;
};

//--- create and place tracks

Variegate.prototype.setState = function(state) {
    EAction.prototype.setState.call(this, state);

    this.setCrosshairCursor();
    this.getDocumentInterface().setClickMode(RAction.PickCoordinate);

    var appWin = RMainWindowQt.getMainWindow();
    var trPos = qsTr("Position");
    this.setCommandPrompt(trPos);
    this.setLeftMouseTip(trPos);
    this.setRightMouseTip(EAction.trCancel);
    EAction.showSnapTools();
};

Variegate.prototype.pickCoordinate = function(event, preview) {
    var di = this.getDocumentInterface();
    var pos = event.getModelPosition();

    this.targetPosition = pos;
    this.targetAngle = 0;

    if (preview) {
      di.previewOperation(this.getOperation(true));
    }
    else {
      var op = this.getOperation(false);
      if (!isNull(op)) {
        di.applyOperation(op);
        di.setRelativeZero(pos);
        this.setState(-1);
        this.terminate();
      }
    }
};

Variegate.prototype.addBlockOperations = function(document, op, preview) {
    var nextFreeId = document.getStorage().getMaxObjectId();
    for (var i = 0; i < this.createBlocks.length; i++) { //create all blocks
      var coa = this.createBlocks[i];
      if (preview && coa[1] != 0) break; //preview only for first line
      var blockId = coa[2];
      var blockReferenceData = new RBlockReferenceData();
      blockReferenceData.setReferencedBlockId(blockId);
      blockReferenceData.setPosition(new RVector(coa[0], coa[1]));
      blockReferenceData.setScaleFactors(new RVector(1.0, 1.0));
      blockReferenceData.setRotation(0.0);
      blockReferenceData.move(this.targetPosition);
      if (this.angle != 0) {
        blockReferenceData.rotate(this.targetAngle, this.targetPosition);
      }
      var blockRef = new RBlockReferenceEntity(document, blockReferenceData);
      op.addObject(blockRef);
    }
};

Variegate.prototype.getOperation = function(preview) {
    var document = this.getDocument();
    var op = new RAddObjectsOperation();

    this.addBlockOperations(document, op, preview);

    for (var i = 0; i < this.createLabels.length; i++) { //create all labels
      var cla = this.createLabels[i];
      if (preview && i > 2) break; //preview only for first three line
      var pos = new RVector(this.targetPosition.getX() + cla[0], this.targetPosition.getY() + cla[1]);
      var text = new RTextEntity(document, new RTextData(
        pos,                 // position
        pos,                 // alignment point
        this.objDistBase,    // height
        2.0,                 // text width (ignored for now)
        RS.VAlignMiddle,     // alignments
        RS.HAlignRight,
        RS.LeftToRight,
        RS.Exact,
        1.0,                 // line spacing factor
        cla[2],       // the text
        "Arial",      // font
        false,        // bold
        false,        // italic
        0.0,          // angle
        false         // simple text without formatting
      ));
      op.addObject(text);
    }
    return op;
};

//--- init

Variegate.init = function(basePath) {
    var action = new RGuiAction(QCTrack.MenuItem.Variegate, RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/Variegate.js");
    action.setGroupSortOrder(QCTrack.MenuOrder.GroupSortOrder);
    action.setSortOrder(QCTrack.MenuOrder.Variegate);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
};

